//
// $Id: gen_filters.cc,v 1.3 2003/05/23 16:03:23 carzanig Exp $
//
#include "ssbconf.h"

#include <sys/time.h>
#include <math.h>

#include <cstdlib>
#include <iostream>
#include <fstream>
#include <algorithm>
#include <set>
#include <string>
#include <sstream>

#include "wset.h"

using namespace std;

void print_usage(const char *name) 
{
    cerr << "usage: " << name << " [options...]\n\
options:\n\
\t -I <num>                       total number of interfaces\n\
\t -N <num>                       total number of notifications generated\n\
\t -seed <num>                    random seed\n\
\n\
\t -fl|--filters-min <num>        number of filters per interface\n\
\t -fh|--filters-max <num>        number of filters per interface\n\
\t -cl|--constr-min <num>         min number of constraints per filter\n\
\t -ch|--constr-max <num>         max number of constraints per filter\n\
\t -al|--attr-min <num>           min number of attributes per notification\n\
\t -ah|--attr-max <num>           max number of attributes per notification\n\
\n\
\t -da|--dist-attributes <file>   distribution of attribute names\n\
                                  default = attr_names.dist\n\
\t -dc|--dist-constraints <file>  distribution of constraint names\n\
                                  default = constr_names.dist\n\
\t -dt|--dist-types <file>        distribution of attribute types\n\
                                  default = types.dist\n\
\t -do|--dist-operators <file>    distribution of operators\n\
                                  default = operators.dist\n\
\t -dso|--dist-string-operators <file>    distribution of string operators\n\
                                  default = string_operators.dist\n\
\t -ds|--dist-str-values <file>   distribution of string values\n\
                                  default = str_values.dist\n\
\t -di|--dist-int-values <file>   distribution of integer values\n\
                                  default = int_values.dist\n\
\t -db|--dist-bool-values <file>  distribution of boolean values\n\
                                  default = bool_values.dist\n\
" << endl;
    exit(1);
}
//
// generator parameters
//
unsigned int		I			= 100;
unsigned int            num                     = 100;
int                     attr_min                = 1;
int                     attr_max                = 10;
int			constr_min		= 1;
int			constr_max		= 5;
int			filters_min		= 1;
int			filters_max		= 20;

//
// we generate an innitial number of filters before we generate any
// notification.  init_filters_count defines this number
//
int			init_filters_count	= 300;
//
// upper bound for active filters.  Filters are added only when
// filters_count < max_filters
//
int			max_filters		= 2000;

//
// lower bound for active filters.  Filters are removed only when
// filters_count > min_filters
//
int			min_filters		= 1000;


const char *		types_dist_f		= "types.dist";
const char *		op_dist_f		= "operators.dist";
const char *		op_string_dist_f	= "string_operators.dist";
const char *		attr_dist_f		= "attr_names.dist";
const char *		constr_dist_f		= "constr_names.dist";
const char *		str_dist_f		= "str_values.dist";
const char *		int_dist_f		= "int_values.dist";
const char *		bool_dist_f		= "bool_values.dist";

long			seed			= 0;
bool			init_seed		= false;

wset<string>		attr_dist;
wset<string>		constr_dist;

wset<string>		op_dist;
wset<string>		op_string_dist;
wset<char>		types_dist;
wset<string>		str_dist;
wset<int>		int_dist;
wset<bool>		bool_dist;


/** random value with uniform distribution in [0,max)
 **/
int rand_normal(int max) {
    return (int) (1.0*max*rand()/(RAND_MAX+1.0));
}

/** random value with uniform distribution in [min,max)
 **/
int rand_normal_range(int min, int max) {
    return min + rand_normal(max - min);
}

unsigned long rand_poisson_delta(unsigned long mean)
{
    double rn = 1.0 - 1.0*rand()/(RAND_MAX+1.0);

    return (unsigned long)(-(1.0*mean)*log(rn));
};

void parse_cmd_line(int argc, char *argv[]) {
    for (int i=1; i< argc; ++i) { 
        if (strcmp(argv[i], "-N")==0) {
            if (++i < argc) {
                num = atoi(argv[i]);
            } else {
                print_usage(argv[0]);
            }
        } else if (strcmp(argv[i], "-I")==0) {
	    if (++i < argc) {
		I = atoi(argv[i]);
	    } else {
		print_usage(argv[0]);
	    }
        } else if (strcmp(argv[i], "-seed")==0) {
	    if (++i < argc) {
		seed = atoi(argv[i]);
		init_seed = true;
	    } else {
		print_usage(argv[0]);
	    }
	} else if (strcmp(argv[i], "-fl")==0 
	    || strcmp(argv[i], "--filters-min")==0) {
	    if (++i < argc) {
		filters_min = atoi(argv[i]);
	    } else {
		print_usage(argv[0]);
	    }
	} else if (strcmp(argv[i], "-fh")==0 
	    || strcmp(argv[i], "--filters-min")==0) {
	    if (++i < argc) {
		filters_max = atoi(argv[i]);
	    } else {
		print_usage(argv[0]);
	    }
	} else if (strcmp(argv[i], "-cl")==0 
	    || strcmp(argv[i], "--constr-min")==0) {
	    if (++i < argc) {
		constr_min = atoi(argv[i]);
	    } else {
		print_usage(argv[0]);
	    }
	} else if (strcmp(argv[i], "-ch")==0
		   || strcmp(argv[i], "--constr-max")==0) {
	    if (++i < argc) {
		constr_max = atoi(argv[i]);
	    } else {
		print_usage(argv[0]);
	    }
        } else if (strcmp(argv[i], "-al")==0
            || strcmp(argv[i], "--attr-min")==0) {
            if (++i < argc) {
                attr_min = atoi(argv[i]);
            } else {
                print_usage(argv[0]);
            }
        } else if (strcmp(argv[i], "-ah")==0
                   || strcmp(argv[i], "--attr-max")==0) {
            if (++i < argc) {
                attr_max = atoi(argv[i]);
            } else {
                print_usage(argv[0]);
            }
	} else if (strcmp(argv[i], "-da")==0
		   || strcmp(argv[i], "--dist-attributes")==0) {
	    if (++i < argc) {
		attr_dist_f = argv[i];
	    } else {
		print_usage(argv[0]);
	    }
	} else if (strcmp(argv[i], "-dc")==0
		   || strcmp(argv[i], "--dist-constraints")==0) {
	    if (++i < argc) {
		constr_dist_f = argv[i];
	    } else {
		print_usage(argv[0]);
	    }
	} else if (strcmp(argv[i], "-do")==0
		   || strcmp(argv[i], "--dist-operators")==0) {
	    if (++i < argc) {
		op_dist_f = argv[i];
	    } else {
		print_usage(argv[0]);
	    }
        } else if (strcmp(argv[i], "-dso") == 0
                   || strcmp(argv[i], "--dist-string-operators") == 0)  {
            if (++i < argc) {
                op_string_dist_f = argv[i];
            } else {
                print_usage(argv[0]);
            }
	} else if (strcmp(argv[i], "-dt")==0
		   || strcmp(argv[i], "--dist-types")==0) {
	    if (++i < argc) {
		types_dist_f = argv[i];
	    } else {
		print_usage(argv[0]);
	    }
	} else if (strcmp(argv[i], "-ds")==0
		   || strcmp(argv[i], "--dist-str-values")==0) {
	    if (++i < argc) {
		str_dist_f = argv[i];
	    } else {
		print_usage(argv[0]);
	    }
	} else if (strcmp(argv[i], "-di")==0
		   || strcmp(argv[i], "--dist-int-values")==0) {
	    if (++i < argc) {
		int_dist_f = argv[i];
	    } else {
		print_usage(argv[0]);
	    }
	} else if (strcmp(argv[i], "-db")==0
		   || strcmp(argv[i], "--dist-bool-values")==0) {
	    if (++i < argc) {
		bool_dist_f = argv[i];
	    } else {
		print_usage(argv[0]);
	    }
	} else {
	    print_usage(argv[0]);
	}
    }
}

ostream & output_attr(ostream & out, 
		      const string & x, 
		      const string & o, 
		      const string & v) {
    //
    // simplistic version
    //
    return out << x << ' ' << o << ' ' << '"' << v << '"';
}

ostream & output_attr(ostream & out, 
		      const string & x, 
		      const string & o, 
		      int v) {
    //
    // simplistic version
    //
    return out << x << ' ' << o << ' ' << v;
}

ostream & output_attr(ostream & out, 
		      const string & x, 
		      const string & o, 
		      bool v) {
    //
    // simplistic version
    //
    return out << x << ' ' << o << ' '<< (v ? "true" : "false");
}

void gen_notif(ostream & out) {
    set<string> attrs;
    int acount = rand_normal_range(attr_min, attr_max);

    while(attrs.size() < acount) 
	attrs.insert(attr_dist.rand_pick());

    set<string>::const_iterator vi;
    for(vi = attrs.begin(); vi != attrs.end(); ++vi) {
	out << ' ';
        char t = types_dist.rand_pick();
        switch(t) {
        case 's': output_attr(out, *vi, "=", str_dist.rand_pick()); break;
        case 'i': output_attr(out, *vi, "=", int_dist.rand_pick()); break;
        case 'b': output_attr(out, *vi, "=", bool_dist.rand_pick()); break;
        default:  out << "???     "; break;
        }
    }
}

void gen_filter(ostream & out) {
    set<string> attrs;
    int acount = rand_normal_range(constr_min, constr_max);

    while(attrs.size() < acount) 
	attrs.insert(attr_dist.rand_pick());

    set<string>::const_iterator vi;

    for(vi = attrs.begin(); vi != attrs.end(); ++vi) {
	if (vi != attrs.begin()) cout << ", ";
	char t = types_dist.rand_pick();
	switch(t) {
	case 's': output_attr(out, *vi, op_string_dist.rand_pick(), 
			      str_dist.rand_pick()); break;
	case 'i': output_attr(out, *vi, op_dist.rand_pick(), 
			      int_dist.rand_pick()); break;
	case 'b': output_attr(out, *vi, op_dist.rand_pick(), 
			      bool_dist.rand_pick()); break;
	default:  out << "???     "; break; 
	}
    }
    // out << endl;
}

int main(int argc, char *argv[]) 
{
    //
    // set parameters from command-line options
    //
    parse_cmd_line(argc, argv);
    
    //
    // initialize random seed
    //
    if (!init_seed) {
	struct timeval tv;
	gettimeofday(&tv, NULL);
	seed = tv.tv_usec;
    }
    srand(seed);

    //
    // read distribution file for strings
    //
    ifstream str_dist_s(str_dist_f);
    if (!str_dist_s) {
	cerr << "error: could not open string values distribution file: " 
	     << str_dist_f << endl;
	return 1;
    }
    str_dist.read_from(str_dist_s);
    str_dist_s.close();
    //
    // read distribution file for integers
    //
    ifstream int_dist_s(int_dist_f);
    if (!int_dist_s) {
	cerr << "error: could not open integer values distribution file: " 
	     << int_dist_f << endl;
	return 1;
    }
    int_dist.read_from(int_dist_s);
    int_dist_s.close();

    //
    // read distribution file for booleans
    //
    ifstream bool_dist_s(bool_dist_f);
    if (!bool_dist_s) {
	cerr << "error: could not open booleans values distribution file: " 
	     << bool_dist_f << endl;
	return 1;
    }
    bool_dist.read_from(bool_dist_s);
    bool_dist_s.close();

    //
    // read distribution file for types
    //
    ifstream types_dist_s(types_dist_f);
    if (!types_dist_s) {
	cerr << "error: could not open types distribution file: " 
	     << types_dist_f << endl;
	return 1;
    }
    types_dist.read_from(types_dist_s);
    types_dist_s.close();

    //
    // read distribution file for operators
    //
    ifstream op_dist_s(op_dist_f);
    if (!op_dist_s) {
	cerr << "error: could not open operators distribution file: " 
	     << op_dist_f << endl;
	return 1;
    }
    op_dist.read_from(op_dist_s);
    op_dist_s.close();

    //
    // for string operators
    //
    ifstream op_string_dist_s(op_string_dist_f);
    if (!op_string_dist_s) {
        cerr << "Error: Could not open string operators distribution file: "
             << op_string_dist_f << endl;
        return 1;
    }
    op_string_dist.read_from(op_string_dist_s);
    op_string_dist_s.close();

    //
    // for string operators
    //
    ifstream attr_dist_s(attr_dist_f);
    if (!attr_dist_s) {
        cerr << "Error: Could not open string operators distribution file: "
             << attr_dist_f << endl;
        return 1;
    }
    attr_dist.read_from(attr_dist_s);
    attr_dist_s.close();

    //
    // for string operators
    //
    ifstream constr_dist_s(constr_dist_f);
    if (!constr_dist_s) {
        cerr << "Error: Could not open string operators distribution file: "
             << constr_dist_f << endl;
        return 1;
    }
    constr_dist.read_from(constr_dist_s);
    constr_dist_s.close();

    //
    // okay, here I need to compute the number of filters per
    // interface, fpi.  I could apply some sophisticated distribution,
    // or better yet, I could supply a distribution file for
    // interfaces.  However for now, I'll just use a Poisson
    // distribution, assuring a mean F / I for fpi, so that for
    // reasonably high numbers of 
    //
    unsigned int i = 0;
    unsigned int f = 0;		// total number of generated filters

    unsigned int notif_count = 0;

    for(i = 0; i < I; ++i) {
	f = rand_normal_range(filters_min, filters_max);
	if (f > 0) {
	    cout << "ifconfig " << i << ' ';
	    gen_filter(cout);
	}
	while(--f > 0) {
	    cout << endl << " | ";
	    gen_filter(cout);
	}
	cout << endl;
    }

    for(i = 0; i < num; ++i) {
	cout << "select ";
	gen_notif(cout);
	cout << endl;
    }

    return 0;
}
